
//***********
//* Includes
//***********
#include <stdio.h>
#include <string.h>


//********************
//* Typendefinitionen
//********************
typedef struct {
  uint8_t     byteCnt;          // Anzahl Bytes im Record
  uint16_t    adr;              // Adresse (HexFile)
  uint8_t     recordType;       // Type des Recors
  uint8_t     record[16];       // Daten im Record
  uint8_t     checksum;         // Checksumme
} sHexFileLine;


//*************
//* Konstanten
//*************
/// Groesse des Speichers in Byte 
#define MemorySize    (256 * 1024)   // 256 kiloByte


//************
//* Variabeln
//************
/// Buffer mit Inhalt des Speichers Byteadressierbar
uint8_t   hexBuffer[MemorySize];

/// Position an welcher das naechste Byte in den Buffer geschrieben wird
/// (wird nach jedem Schreibzugriff inkrementiert)
uint32_t  fillPos;

/// Buffer fr eine Zeile im HexFile (ASCII-Text)
char      fileLine[100];


//********************************
//* Lokale Funktionsdeklarationen
//********************************
static void     writeLineToFile   ( FILE *file, sHexFileLine *data );
static uint8_t  getChecksum       ( const sHexFileLine *data );
static uint8_t  getLowByte        ( const uint16_t num );
static uint8_t  getHighByte       ( const uint16_t num );
static void     uint2hex          ( char *string, const uint16_t num, const uint8_t bytes );


//*************
//* Funktionen
//*************
//----------------------------------------------------------------------
void clearBuffer ( uint8_t fillValue ) {
  uint32_t  i;
  
  fillPos = 0;
  for (i=0; i<MemorySize; i++) {
    hexBuffer[i] = fillValue;
  }
}


//----------------------------------------------------------------------
void writeBuffer ( uint8_t value ) {
  hexBuffer[fillPos] = value;
  fillPos++;
}


//----------------------------------------------------------------------
void writeBufferAdress ( uint32_t byteAdress, uint8_t value ) {
  hexBuffer[byteAdress] = value;
}


//----------------------------------------------------------------------
void setBufferTop ( uint32_t topAdress ) {
  fillPos = topAdress + 1;
}


//----------------------------------------------------------------------
int generateHexFile ( const char *filename ) {
  FILE          *pFile;
  uint32_t      absoluteAdr;    // Adresse im Buffer
  uint32_t      relativeAdr;    // Adresse im HexFile (16bit)
  uint16_t      segmentAdr;     // Segment Adresse im HexFile
  sHexFileLine  line;           // Datenstruktur fuer eine Zeile
  uint8_t       i;
  
  // Leere Datei zum schreiben erstellen
  pFile = fopen(filename, "w");
  
  if (pFile == NULL) {
    printf("Fehler beim erstellen der Datei");
    fclose(pFile);
    return 1;
  }
  
  relativeAdr = 0;
  for (absoluteAdr=0; absoluteAdr<fillPos; absoluteAdr+=16) {
    if (relativeAdr > 0xFFFF) {
      // "Extended Segment Address Record"
      // ---------------------------------
      relativeAdr = 0;
      segmentAdr = (absoluteAdr - 1) / 16 + 1;
      
      line.byteCnt    = 0x02;
      line.adr        = 0x0000;
      line.recordType = 0x02;
      line.record[0]  = getHighByte(segmentAdr);
      line.record[1]  = getLowByte(segmentAdr);
      line.checksum   = getChecksum(&line);
      
      writeLineToFile(pFile, &line);
    }
    
    // "Data Record"
    // -------------
    if ((fillPos - absoluteAdr) >= 16) {
      line.byteCnt = 16;
    } else {
      line.byteCnt = fillPos - absoluteAdr - 1;
    }
    
    line.adr          = relativeAdr;
    line.recordType   = 0x00;
    
    for (i=0; i<line.byteCnt; i++) {
      line.record[i] = hexBuffer[absoluteAdr + i];
    }
    
    line.checksum     = getChecksum(&line);
    
    writeLineToFile(pFile, &line);
    
    relativeAdr += 16;
  }
  
  // "End of File Record"
  // --------------------
  line.byteCnt      = 0;
  line.adr          = 0x0000;
  line.recordType   = 0x01;
  line.checksum     = 0xFF;
  
  writeLineToFile(pFile, &line);
  
  fclose(pFile);
  return 0;
}


//********************
//* Lokale Funktionen
//********************
//----------------------------------------------------------------------
/// \brief
/// Konvertiert die Rohdaten eine Zeile in ASCII-Zeichen und schreibt
/// dann eine komplette Zeile in das HexFile
/// \param
/// file Pointer auf eine geoeffnete Datei
/// \param
/// data Pointer auf die Zeileninformationen
static void writeLineToFile ( FILE *file, sHexFileLine *data ) {
  char      line[100];
  char      buffer[10];
  uint8_t   i;

  sprintf(line, ":");
  
  uint2hex(buffer, data->byteCnt, 1);
  strcat(line, buffer);
  
  uint2hex(buffer, data->adr, 2);
  strcat(line, buffer);
  
  uint2hex(buffer, data->recordType, 1);
  strcat(line, buffer);
  
  if (data->byteCnt != 0) {
    for (i=0; i<(data->byteCnt); i++) {
      uint2hex(buffer, data->record[i], 1);
      strcat(line, buffer);
    }
  }
  
  uint2hex(buffer, data->checksum, 1);
  strcat(line, buffer);
  strcat(line, "\r\n");
  
  fputs(line, file);
  
}


//----------------------------------------------------------------------
/// \brief
/// Berechent die Checksumme des DataRecords einer Zeile
/// \param
/// data Pointer auf die Zeileninformationen
/// \return
/// Checksumme als Byte
static uint8_t getChecksum ( const sHexFileLine *data ) {
  uint16_t  sum = 0;
  uint8_t   cs;
  uint8_t   i;
  
  sum += data->byteCnt;
  sum += getLowByte(data->adr);
  sum += getHighByte(data->adr);
  sum += data->recordType;
  
  for (i=0; i<data->byteCnt; i++) {
    sum += data->record[i];
  }
  
  cs = (uint8_t)sum & 0xFF;   // Nur LowByte verwenden
  cs = (cs ^ 0xFF) + 1;       // Zweierkomplement
  
  return (uint8_t)cs;
}


//----------------------------------------------------------------------
/// \brief
/// Konvertiert einen Integer in einen String in Hexadezimaler darstellung
/// \param
/// string Pointer auf den String in den geschrieben werden soll
/// \param
/// num Zahl als 16bit Integer
/// \param
/// bytes Anzahl Bytes (1 oder 2)
static void uint2hex ( char *string, const uint16_t num, const uint8_t bytes ) {
  static const char HexDigits[] = {'0', '1', '2','3', '4', '5', '6', '7',
                                   '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
  
  uint8_t   digit;
  uint16_t  zahl    = num;
  
  if (bytes == 1) {
    strcpy(string, "00");
    string += 1;
  } else {
    strcpy(string, "0000");
    string += 3;
  }
  
  do {
    digit = zahl % 16;
    zahl /= 16;
    *string-- = HexDigits[digit];
  } while (zahl > 0);
}


//----------------------------------------------------------------------
static uint8_t getLowByte ( const uint16_t num ) {
  return (uint8_t)(num & 0xFF);
}


//----------------------------------------------------------------------
static uint8_t getHighByte ( const uint16_t num ) {
  return (uint8_t)((num & 0xFF00) >> 8);
}
